#include "tsynctest.h"
#include "example.h"

static void usage( char *pgmName );
static void printFuncs();
static void printSyntax(char* pcFunctionName);
static TSYNC_ERROR GetValuePositions(char *pgmName, char* pcExpression, int* pcValuePositions);
static void evaluate(TSYNC_BoardHandle hnd, char *pgmName, char* pcExpression);
static void evalFile(TSYNC_BoardHandle hnd, char *pgmName, char* pcFileName);

#define OPT_ARG_LIST        "hd:i:e:ps:f:"

int main (int argc, char *argv[])
{
    TSYNC_BoardHandle   hnd;
    int             rc;
    int             ret = EXIT_SUCCESS;
    int             devIdx = DEFAULT_DEV_NUM;
    char            *devName = DEFAULT_DEV;
    int             optRc;
    char            fullDevName[32];
    char            *pgmName = argv[0];
    char            funcNameBuff[50];
    char            cInput[TEST_MAX_EXPRESSION_LENGTH];
    int             bExpression = 0;
    int             bFile = 0;

    opterr = 0;
    while ( (optRc = getopt (argc, argv, OPT_ARG_LIST)) != -1 )
    {
        switch ( optRc )
        {
            case 'd' :
                devName = optarg;
                break;
            case 'i' :
                devIdx = atoi(optarg);
                break;
            case 'e' :
                memset(cInput, '\0', sizeof(cInput));
                sprintf(cInput, optarg, sizeof(optarg));
                bExpression = 1;
                break;
            case 'f' :
                memset(cInput, '\0', sizeof(cInput));
                sprintf(cInput, optarg, sizeof(optarg));
                bFile = 1;
                break;
            case 'p' :
                printFuncs();
                return EXIT_FAILURE;
            case 's' :
                memset(cInput, '\0', sizeof(cInput));
                sprintf(cInput, optarg, sizeof(optarg));
                sscanf(cInput, "%49s", funcNameBuff);
                printSyntax(funcNameBuff);
                return EXIT_FAILURE;
            case 'h' :
            default:
                usage( pgmName );
                return EXIT_FAILURE;
        }
    }

    /*
     * If leftover args, then command is misunderstood...
     * so let user know.
     */
    if ( optind != argc )
    {
        usage( pgmName );
        return EXIT_FAILURE;
    }

    /*
     * construct the device name
     */
    (void)sprintf( fullDevName, "%s%s%d", DEVICE_ROOT, devName, devIdx );

    /*
     * open the device
     */
    if ( (rc = TSYNC_open(&hnd, fullDevName)) != TSYNC_SUCCESS )
    {
        /*
         * if errors, close it
         */
        printf("%s: !Could not open <%s> : rc <%d>\n",
                        pgmName, fullDevName, rc);
        return EXIT_FAILURE;
    }
    
    if(bExpression)
    {
        evaluate(hnd, pgmName, cInput);
    }
    else if(bFile)
    {
        evalFile(hnd, pgmName, cInput);
    }
    
    /*
     * close the handle to the board
     */
    if ( (rc = TSYNC_close(hnd)) != TSYNC_SUCCESS )
    {
        (void) printf("%s: error closing <%s> rc: <%d>\n",
                pgmName, fullDevName, rc );
        ret = 1;
    }

    return (ret);
}

void evalFile(TSYNC_BoardHandle hnd, char *pgmName, char* pcFileName)
{
    char cExpression[TEST_MAX_EXPRESSION_LENGTH] = "";
    char cFirstArg[TEST_MAX_EXPRESSION_LENGTH] = "";
    int nSleepDuration = 0;
    
    // Make sure we can open the file and read from it
    FILE *fp;
    if(NULL == (fp = fopen(pcFileName, "r")))
    {
        printf("Could not open image file for reading: %s\n", pcFileName);
        return;
    }
    
    int i;
    for(i = 0; i < 256; i ++)
    {
        int rc = fscanf(fp, "%256[^\n]\n", cExpression);

        if(rc == EOF || rc == 0) break;
        if(cExpression[0] == ';') continue;

        printf(" ** Line %u: '%s' **\n", i, cExpression);
        printf("\n");
        
        rc = sscanf(cExpression, "%256s %10d", cFirstArg, &nSleepDuration);
        if(rc == 2 && strcmp(cFirstArg, "sleep") == 0)
        {
            printf(" Sleeping for %d second%s...\n\n",
                nSleepDuration, nSleepDuration == 1 ? "" : "s");
            TSYNC_SLEEP(nSleepDuration);
            continue;
        }
        
        evaluate(hnd, pgmName, cExpression);
        
        printf("\n");
    }
    
    //Close the file handle
    fclose(fp);
}
void evaluate(TSYNC_BoardHandle hnd, char *pgmName, char* pcExpression)
{
    char            funcNameBuff[50];
    TSYNC_ERROR err = TSYNC_SUCCESS;
    sscanf(pcExpression, "%49s", funcNameBuff);
    
    int nValuePositions[TEST_MAX_EXPRESSION_VALUES];
    int i;
    for(i = 0; i < TEST_MAX_EXPRESSION_VALUES; i ++)
        nValuePositions[i] = TEST_INVALID_VALUE_INDEX;
    if(GetValuePositions(pgmName, pcExpression, nValuePositions) != TSYNC_SUCCESS)
    {
        usage( pgmName );
        return;
    }

    int bFoundFunc = 0;
    int bErr = 0;
    
    //allows us to break out of the following section on error
    int loopVar;
    for(loopVar = 0; loopVar < 1; loopVar++)
    {

#define TSYNC_X_DEFINITION_ACQ(fieldType, fieldName) \
    fieldType fieldName;
#define TSYNC_X_DEFINITION_DSP(fieldType, fieldName) ;

#define TSYNC_X_EXP_VALUE_ACQ(fieldName, printType, desc, funcAcc, valueSource) \
    if( ((valueSource & funcAcc) == funcAcc)) /* if equal, source is user */ { \
        if(nValuePositions[pos] == TEST_INVALID_VALUE_INDEX) \
            { bErr = 1; break; } /* too few parameters were found */ \
        int ret = sscanf(pcExpression + nValuePositions[pos++], printType, &fieldName); \
        if(ret != 1) { bErr = 1; break; } /* no value scanned in (incorrect type) */ \
        printf("  " desc ": " printType "\n", fieldName); }
#define TSYNC_X_EXP_VALUE_DSP(fieldName, printType, desc, funcAcc, valueSource) \
    if( ((valueSource & funcAcc) != funcAcc)) /* if not equal, source is board */ { \
        printf("  " desc ": " printType "\n", fieldName); }
    
#define TSYNC_X_EXP_VALUE_S_ACQ(fieldName, printType, desc, funcAcc, valueSource) \
    if( ((valueSource & funcAcc) == funcAcc)) /* if equal, source is user */ { \
        if(nValuePositions[pos] == TEST_INVALID_VALUE_INDEX) \
            { bErr = 1; break; } /* too few parameters were found */ \
        int ret = sscanf(pcExpression + nValuePositions[pos++], printType, fieldName); \
        if(ret != 1) { bErr = 1; break; } /* no value scanned in (incorrect type) */ \
        printf("  " desc ": " printType "\n", fieldName); }
#define TSYNC_X_EXP_VALUE_S_DSP(fieldName, printType, desc, funcAcc, valueSource) \
    if( ((valueSource & funcAcc) != funcAcc)) /* if not equal, source is board */ { \
        printf("  " desc ": " printType "\n", fieldName); }
    
#define TSYNC_X_EXP_VALUE_D_ACQ(fieldType, fieldName, printType, desc, funcAcc, valueSource) \
    TSYNC_X_DEFINITION_ACQ(fieldType, fieldName) \
    TSYNC_X_EXP_VALUE_ACQ(fieldName, printType, desc, funcAcc, valueSource)
#define TSYNC_X_EXP_VALUE_D_DSP(fieldType, fieldName, printType, desc, funcAcc, valueSource) \
    TSYNC_X_EXP_VALUE_DSP(fieldName, printType, desc, funcAcc, valueSource)
    
#define TSYNC_X_UNSP(functionName) \
     \
    if(strcmp(#functionName, funcNameBuff) == 0) \
    { \
        bFoundFunc = 1; \
        \
        printf("  ** Unsupported: "#functionName"\n"); \
    } \

#define TSYNC_X_FUNC(functionName, funcExpVals, customOutput, funcAcc, functionDesc, ...) \
     TSYNC_X_FUNC_START(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     TSYNC_X_FUNC_MID(functionName, funcExpVals, customOutput, funcAcc, functionDesc, __VA_ARGS__) \
     TSYNC_X_FUNC_END(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
    
#define TSYNC_X_FUNC2(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     TSYNC_X_FUNC_START(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     TSYNC_X_FUNC_MID2(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     TSYNC_X_FUNC_END(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
    
#define TSYNC_X_FUNC_START(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     \
    if(strcmp(#functionName, funcNameBuff) == 0) \
    { \
        bFoundFunc = 1; \
        \
        printf(" " functionDesc "\n"); /*printf(" " functionDesc ": %s\n", pcExpression);*/ \
        \
        int pos = 1; \
        TEST_##funcExpVals##_EXP_VALUES(ACQ, funcAcc) \
        if(nValuePositions[pos] != TEST_INVALID_VALUE_INDEX) \
            { bErr = 1; break; } /* too many parameters were found */ \
        \
    
#define TSYNC_X_FUNC_MID(functionName, funcExpVals, customOutput, funcAcc, functionDesc, ...) \
        err = TSYNC_##functionName(hnd, __VA_ARGS__); \
    
#define TSYNC_X_FUNC_MID2(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
        err = TSYNC_##functionName(hnd); \
    
#define TSYNC_X_FUNC_END(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     \
        if (err == TSYNC_SUCCESS) { printf("  ** Success!\n"); \
            TEST_##funcExpVals##_EXP_VALUES(DSP, funcAcc) \
            customOutput \
        } \
        \
        else { printf("  ** Error: %s.\n", tsync_strerror(err)); } \
    } \
    
    TEST_FUNCTION_FIELDS
#undef TSYNC_X_FUNC2
#undef TSYNC_X_FUNC
#undef TSYNC_X_UNSP
#undef TSYNC_X_EXP_VALUE_S_DSP
#undef TSYNC_X_EXP_VALUE_S_ACQ
#undef TSYNC_X_EXP_VALUE_D_DSP
#undef TSYNC_X_EXP_VALUE_D_ACQ
#undef TSYNC_X_EXP_VALUE_DSP
#undef TSYNC_X_EXP_VALUE_ACQ
#undef TSYNC_X_DEFINITION_DSP
#undef TSYNC_X_DEFINITION_ACQ
    
    }
    
    if(!bFoundFunc)

    {
        printf(" '%s' is unknown.  Use -p to see a list of known functions.\n", funcNameBuff);
    }
    if(bErr)
    {
        printf(" ** Syntax Error **\n");
        printSyntax(funcNameBuff);
    }
}

TSYNC_ERROR GetValuePositions(char *pgmName, char* pcExpression, int* pcValuePositions)
{
    int numValuesFound = 1;
    int expIndex = 0;
    while(pcExpression[expIndex] != '\0')
    {
        if(pcExpression[expIndex] == 32) //found a space!
        {
            int nextValIndex = expIndex + 1;
            if(pcExpression[nextValIndex] == '\0')
            {
                return TSYNC_SUCCESS;
            }
            else if(pcExpression[nextValIndex] == 32)
            {
                expIndex ++;
                continue;
            }
            
            pcValuePositions[numValuesFound] = nextValIndex;
            numValuesFound ++;
        }
        expIndex ++;
    }
    return TSYNC_SUCCESS;
}

void printFuncs()
{
    printf(" Known functions:\n");
    
#define TSYNC_X_UNSP(functionName) \
    printf("  **Unsupported: "#functionName"\n");
    
#define TSYNC_X_FUNC(functionName, funcExpVals, customOutput, funcAcc, functionDesc, ...) \
    printf("  "#functionName" ("#functionDesc")""\n");
    
#define TSYNC_X_FUNC2(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
    printf("  "#functionName" ("#functionDesc")""\n");
    
        TEST_FUNCTION_FIELDS
        
#undef TSYNC_X_FUNC2
#undef TSYNC_X_FUNC
#undef TSYNC_X_UNSP
}

void printSyntax(char* pcFunctionName)
{
    int bFoundFunc = 0;
    int bFoundParam = 0;
    
#define TSYNC_X_DEFINITION_DSP(fieldType, fieldName) ;

#define TSYNC_X_EXP_VALUE_DSP(fieldName, printType, desc, funcAcc, valueSource) \
    if( ((valueSource & funcAcc) == funcAcc)) /* if equal, source is user */ { \
    printf("  " desc ": %" printType "\n"); \
    bFoundParam = 1; }
    
#define TSYNC_X_EXP_VALUE_S_DSP(fieldName, printType, desc, funcAcc, valueSource) \
    TSYNC_X_EXP_VALUE_DSP(fieldName, printType, desc, funcAcc, valueSource)
    
#define TSYNC_X_EXP_VALUE_D_DSP(fieldType, fieldName, printType, desc, funcAcc, valueSource) \
    TSYNC_X_EXP_VALUE_DSP(fieldName, printType, desc, funcAcc, valueSource)
    
#define TSYNC_X_UNSP(functionName) \
     \
    if(strcmp(#functionName, pcFunctionName) == 0) \
    { \
        bFoundFunc = 1; \
        bFoundParam = 1; \
        \
        printf("  Unsupported: "#functionName"\n"); \
    } \
    
#define TSYNC_X_FUNC(functionName, funcExpVals, customOutput, funcAcc, functionDesc, ...) \
    TSYNC_X_FUNC2(functionName, funcExpVals, customOutput, funcAcc, functionDesc)
    
#define TSYNC_X_FUNC2(functionName, funcExpVals, customOutput, funcAcc, functionDesc) \
     \
    if(strcmp(#functionName, pcFunctionName) == 0) \
    { \
        bFoundFunc = 1; \
        \
        printf(" Syntax for: " #functionName " ("functionDesc")\n"); /*printf(" " functionDesc ": %s\n", pcExpression);*/ \
        \
        TEST_##funcExpVals##_EXP_VALUES(DSP, funcAcc) \
    } \
    
    TEST_FUNCTION_FIELDS
#undef TSYNC_X_FUNC2
#undef TSYNC_X_FUNC
#undef TSYNC_X_UNSP
#undef TSYNC_X_EXP_VALUE_S_DSP
#undef TSYNC_X_EXP_VALUE_D_DSP
#undef TSYNC_X_EXP_VALUE_DSP
#undef TSYNC_X_DEFINITION_DSP
    
    if(!bFoundFunc)
    {
        printf("  '%s' is unknown.  Use -p to see a list of known functions.\n", pcFunctionName);
    }
    else if(!bFoundParam)
    {
        printf("  No input paramaters.\n");
    }
    
}

void usage( char *pgmName )
{
    (void)printf("Usage: %s \n"
                 "          [-h] [-d NAME] [-i INDEX] [-p] [-s FUNCNAME]\n"
                 "          [-e \"FUNCNAME arg1 arg2 ..\"] [-f FILENAME]\n\n", pgmName );
    (void)printf("\t-h   help message\n");
    (void)printf("\t-d   device name (e.g. %s)\n", DEFAULT_DEV);
    (void)printf("\t-i   device index (e.g. 0,1,..)\n");
    (void)printf("\t\tsuch that combined forms: %s0\n", DEFAULT_DEV);
    (void)printf("\t-e   execute library function\n");
    (void)printf("\t-p   print available functions\n");
    (void)printf("\t-s   print syntax for FUNCNAME\n");
    (void)printf("\t-f   execute the file FILENAME\n");
    (void)printf("\n");
}
