#include "example.h"

static void usage( char *pgmName );

#define OPT_ARG_LIST        "hd:i:f:"

static
uint8_t WriteWordEndianSafe(uint8_t* pDst, uint8_t* pSrc, unsigned int wordLength);

void PrintBuffer(
        uint8_t*    pBuffer,
        uint32_t    ulLen)
{
    /* Print the serialized contents of the buffer */
    unsigned int i;
    printf("\n*******\n");
    for(i = 0; i < ulLen; i++)
    {
        if(i % 16 == 0)
        {
            printf("0x%02X: ", i);
        }

        printf("%02X ", pBuffer[i]);

        if(i % 16 == 7)
        {
            printf(" ");
        }
        else if(i % 16 == 15)
        {
            printf("\n");
        }
        
    }
    printf("\n*******\n\n");
}

int main (int argc, char *argv[])
{
    const int MAX_FILENAME_LENGTH = 1024;
    TSYNC_BoardHandle   hnd;
    int             rc;
    int             ret = 0;
    int             devIdx = DEFAULT_DEV_NUM;
    char            *devName = DEFAULT_DEV;
    int             optRc;
    char            fullDevName[MAX_FILENAME_LENGTH];
    char            *pgmname = argv[0];
    char            cFileName[MAX_FILENAME_LENGTH];

    opterr = 0;
    while ( (optRc = getopt (argc, argv, OPT_ARG_LIST)) != -1 )
    {
        switch ( optRc )
        {
            case 'd' :
                devName = optarg;
                break;
            case 'i' :
                devIdx = atoi(optarg);
                break;
            case 'f' :
                TSYNC_SAFE_SPRINTF( cFileName, MAX_FILENAME_LENGTH, "%s", optarg );
                break;
            case 'h' :
            default:
                usage( pgmname );
                return (1);
        }
    }

    /*
     * If leftover args, then command is misunderstood...
     * so let user know.
     */
    if ( optind != argc/* || strcmp(cFileName, DEFAULT_IMAGE_FILE) == 0*/)
    {
        usage( pgmname );
        return (1);
    }

    /*
     * construct the device name
     */
    (void)sprintf( fullDevName, "%s%s%d", DEVICE_ROOT, devName, devIdx );

    /*
     * open the device
     */
    if ( (rc = TSYNC_open(&hnd, fullDevName)) != TSYNC_SUCCESS )
    {
        /*
         * if errors, close it
         */
        printf("%s: !Could not open <%s> : rc <%d>\n",
                        pgmname, fullDevName, rc);
        return (1);
    }
    
    // Make sure we can open the file and read from it
    FILE *fImage;
    if(NULL == (fImage = fopen(cFileName, "rb")))
    {
        printf("Could not open image file for reading: %s\n", cFileName);
        return (1);
    }
    
    //init
    TSYNC_ULImageHeaderObj header;
    TSYNC_UpdateEndObj trailer;
    
    //Read from the file to get header info.
    {
        printf(" Reading Header:\n");
        unsigned char tmp[4];
        int sizeDst = 0;
        
        //get header.mark out of file
        sizeDst = sizeof(header.mark);
        if(0 != fread(&tmp, 1, sizeDst, fImage))
        {
            //copy to destination, swapping if needed
            WriteWordEndianSafe((uint8_t*)&header.mark, tmp, sizeDst);
            
            printf("  header.mark: 0x%08x\n", header.mark);
        }
        else
        {
            printf("  Could not read header.mark\n");
            fclose(fImage);
            return (1);
        }
        
        //get header.type out of file
        sizeDst = sizeof(header.type);
        if(0 != fread(&tmp, 1, sizeDst, fImage))
        {
            //copy to destination, swapping if needed
            WriteWordEndianSafe((uint8_t*)&header.type, tmp, sizeDst);
            
            printf("  header.type: 0x%08x\n", header.type);
        }
        else
        {
            printf("  Could not read header.type\n");
            fclose(fImage);
            return (1);
        }
        
        //get header.len out of file
        sizeDst = sizeof(header.len);
        if(0 != fread(&tmp, 1, sizeDst, fImage))
        {
            //copy to destination, swapping if needed
            WriteWordEndianSafe((uint8_t*)&header.len, tmp, sizeDst);
            
            printf("  header.len: 0x%08x (%u)\n", header.len, header.len);
        }
        else
        {
            printf("  Could not read header.len\n");
            fclose(fImage);
            return (1);
        }
        
        printf(" Reading Trailer:\n");
        fseek(fImage, -8, SEEK_END);
        
        trailer.type = header.type;
        
        //get trailer.ver out of file
        sizeDst = sizeof(trailer.ver);
        if(0 != fread(&tmp, 1, sizeDst, fImage))
        {
            //copy to destination, swapping if needed
            //WriteWordEndianSafe((uint8_t*)&trailer.ver, tmp, sizeDst);
            memcpy(&trailer.ver, tmp, sizeDst);
#define EXPECTED_VERSION_LENGTH 4
            
            char verBuff[EXPECTED_VERSION_LENGTH + 1];
            memset(verBuff, 0, EXPECTED_VERSION_LENGTH + 1);
            strncpy(verBuff, (char*)trailer.ver, EXPECTED_VERSION_LENGTH);
            
            printf("  trailer.ver: %s\n", verBuff);
        }
        else
        {
            printf("  Could not read trailer.ver\n");
            fclose(fImage);
            return (1);
        }
        
        //get trailer.crc out of file
        sizeDst = sizeof(trailer.crc);
        if(0 != fread(&tmp, 1, sizeDst, fImage))
        {
            //copy to destination, swapping if needed
            WriteWordEndianSafe((uint8_t*)&trailer.crc, tmp, sizeDst);
            
            printf("  trailer.crc: 0x%08x\n", trailer.crc);
        }
        else
        {
            printf("  Could not read trailer.crc\n");
            fclose(fImage);
            return (1);
        }
        
    }
    
    //Cancel Update
    {
        printf(" Canceling any previously started update . . . \n");
        TSYNC_ERROR err = TSYNC_US_cancel(hnd, header.type);
        
        if (err == TSYNC_SUCCESS)
        {
            printf(" Cancel Update: Success.\n");
        }
        else
        {
            printf(" Cancel Update error: %s.\n", tsync_strerror(err));
        }
    }
    
    //Start Update
    {
        TSYNC_ERROR err = TSYNC_US_start(hnd, &header);
        
        if (err == TSYNC_SUCCESS)
        {
            printf(" Start Update: Success.\n");
        }
        else
        {
            printf(" Start Update error: %s.\n", tsync_strerror(err));
        }
    }
    
    //Do the Update
    {
        unsigned char buff[TSYNC_DATA_BLOCK_SIZE];
        memset(buff, '\0', TSYNC_DATA_BLOCK_SIZE);
        
        fseek(fImage, 0, SEEK_SET);
        int i = 0;
        int nTotalRead = 0;
        
        int nMaxIters = header.len / TSYNC_DATA_BLOCK_SIZE + 1;
        
        for(i = 0; i < nMaxIters; i++)
        {
            TSYNC_UpdateDataObj pObj;
            pObj.type = header.type;
            
            memset(pObj.data, '\0', TSYNC_DATA_BLOCK_SIZE);
            int nBytesRead = fread(pObj.data, 1, TSYNC_DATA_BLOCK_SIZE, fImage);
            if(nBytesRead == 0) break;
            nTotalRead += nBytesRead;
            
            TSYNC_ERROR err = TSYNC_US_data(hnd, &pObj);
            
            if (err == TSYNC_SUCCESS)
            {
                printf(".");

#define DOTS_PER_LINE 50

                if ((i % DOTS_PER_LINE == (DOTS_PER_LINE-1)) || 
                   (i == nMaxIters - 1))
                {
                    printf(" %7u bytes (%3u%%)\n", nTotalRead, 
                           ((i+1)*100)/nMaxIters);
                }
            }
            else
            {
                printf(" Do Update error: %s.\n", tsync_strerror(err));
                break;
            }
        }
    }
    
    //End Update
    {
        TSYNC_ERROR err = TSYNC_US_end(hnd, &trailer);
        
        if (err == TSYNC_SUCCESS)
        {
            printf(" End Update: Success.\n");
        }
        else
        {
            printf(" End Update error: %s.\n", tsync_strerror(err));
        }
    }
    
    //Cancel Update
    /*{
        TSYNC_ERROR err = TSYNC_US_cancel(hnd, header.type);
        
        if (err == TSYNC_SUCCESS)
        {
            printf(" Cancel Update: Success.\n");
        }
        else
        {
            printf(" Cancel Update error: %s.\n", tsync_strerror(err));
        }
    }*/
    
    //Close the file handle
    fclose(fImage);
    
    /*
     * close the handle to the board
     */
    if ( (rc = TSYNC_close(hnd)) != TSYNC_SUCCESS )
    {
        (void) printf("%s: error closing <%s> rc: <%d>\n",
                pgmname, fullDevName, rc );
        ret = 1;
    }

    return (ret);
}

uint8_t WriteWordEndianSafe(uint8_t* pDst, uint8_t* pSrc, unsigned int wordLength) 
{

#ifdef HOST_BIG_ENDIAN
    memcpy(pDst, pSrc, wordLength);
#else
    uint8_t iByte;
    for(iByte = 0; iByte < wordLength; iByte++)
    {
        pDst[iByte] = pSrc[wordLength - iByte - 1];
    }
#endif

    return wordLength;
}

void usage( char *pgmName )
{
    (void)printf("Usage: %s [-h] [-d NAME] [-i INDEX] \n\n", pgmName );
    (void)printf("\t-h   help message\n");
    (void)printf("\t-d   device name (e.g. %s)\n", DEFAULT_DEV);
    (void)printf("\t-i   device index (e.g. 0,1,..)\n");
    (void)printf("\t\tsuch that combined forms: %s0\n", DEFAULT_DEV);
    (void)printf("\t-f   file to download\n");
    (void)printf("\n");
}
